import logging
import time
import datetime
import os
import glob
from wb_common import beeper
from ast import literal_eval
from wb_modbus import minimalmodbus
from argparse import ArgumentParser


BACKUP_DIR = '/root/ir_backup'


if os.path.exists(BACKUP_DIR):
    pass
else:
    os.mkdir(BACKUP_DIR)


def get_default_parser():
    parser = ArgumentParser(add_help=True)
    parser.add_argument('-a', '--slave', dest='slaveid', type=int, help='Modbus ID of AC monitor to calibrate', required=True)
    parser.add_argument('--port', dest='port', type=str, help='Select port to use', default='/dev/ttyRS485-2')
    parser.add_argument('--uart-params', dest='uart_params_str', type=str, help="Testing device's serial port settings (will be set on calibration and used in other cases)", default='9600N2')
    return parser


class IRMixin(object):
    """
    Sending IR-signal (written into referece device) between reference and testing.
    Reading IR-frames durations from modbus regs (look into wiki for IR-cmd format info).
    """
    MAX_NUMBER_OF_BANKS = 80

    # Base regs for each rom bank (base + bank_number)
    EDIT_BANK_BASE = 5200
    BANK_BYTELENGTH_BASE = 5400

    # Common regs for all rom banks
    LEARN_TO_RAM_COIL = 5001
    RAM_BOUNDS = [2000, 2509]

    # Implementation params
    _SERIAL_TIMEOUT = 3.0  # Reading many regs requires long timeout
    # _SCHEDULER_PERIOD = 200E-3  # Look into ms firmware. 20ms with a huge reserv
    _SCHEDULER_PERIOD = 1.0  # Look into ms firmware. 20ms with a huge reserv

    def __init__(self):
        self.BACKUP_FNAME = os.path.join(BACKUP_DIR, '%%d_%s.txt' % datetime.datetime.now().strftime("%m%d%Y_%H:%M:%S"))
        self.BACKUP_FNAME.replace('%%', '%')

    @classmethod
    def set_serial_timeout(cls, instrument, serial_timeout):
        setattr(instrument, '_initial_timeout', instrument.device.serial.timeout)  # backing up initial serial timeout
        instrument.device.serial.timeout = serial_timeout

    @classmethod
    def restore_serial_timeout(cls, instrument):
        instrument.device.serial.timeout = getattr(instrument, '_initial_timeout')  # restoring initial serial timeout

    @classmethod
    def get_used_banks(cls, instrument):
        logging.debug("Will poll IR banks to check for used and empty")
        used, empty = [], []
        for bank, reg in enumerate(range(cls.BANK_BYTELENGTH_BASE, cls.BANK_BYTELENGTH_BASE + cls.MAX_NUMBER_OF_BANKS), 0):
            try:
                regs_length = instrument.read_u16(reg)
                logging.debug("Bank %d has %d ir regs" % (bank, regs_length))
                if regs_length == 0:
                    empty.append(bank)
                else:
                    used.append(bank)
            except minimalmodbus.IllegalRequestError as e:
                logging.warning("Bank %d does not exist!" % bank)
                break
        return used, empty

    @classmethod
    def calculate_cmd_duration(cls, values_row):
        return sum(values_row) * 10E-6  # values are 10us-quants

    @classmethod
    def clear_ram(cls, instrument, delay=2.0):
        if cls.read_ram(instrument, 2) != [0, 0]:  #  empty 2 first regs in a RAM mean empty RAM
            logging.debug("Clearing ram by learning empty signal...")
            instrument.write_bit(cls.LEARN_TO_RAM_COIL, True)
            time.sleep(delay)
            instrument.write_bit(cls.LEARN_TO_RAM_COIL, False)
            time.sleep(1.0)

    @classmethod
    def read_ram(cls, instrument, regs_length=0):
        """
        Reading a sequence of holding regs from RAM (RAM_BOUNDS[1] - RAM_BOUNDS[0])
        """
        if regs_length > cls.RAM_BOUNDS[1] - cls.RAM_BOUNDS[0]:
            raise RuntimeError("Too long sequence! (should not be more %d)" % cls.RAM_BOUNDS[1] - cls.RAM_BOUNDS[0])
        ret = []
        logging.debug("Reading signal from RAM as %d regs with %.2fs serial timeout..." % (regs_length, cls._SERIAL_TIMEOUT))
        cls.set_serial_timeout(instrument, cls._SERIAL_TIMEOUT)
        _max_len = 125  # for modbus funcode 3
        if regs_length > _max_len:  # will read _max_len regs once
            ret.extend(instrument.read_u16_holdings(cls.RAM_BOUNDS[0], _max_len))
            _beginning = cls.RAM_BOUNDS[0] + _max_len
            while _beginning - cls.RAM_BOUNDS[0] < regs_length:  # and per 1 regs further
                ret.append(instrument.read_u16(_beginning))
                _beginning += 1
        elif regs_length == 0:
            logging.warning("Will not read signal")
        else:
            ret = instrument.read_u16_holdings(cls.RAM_BOUNDS[0], regs_length)
        logging.debug("RAM: %s" % str(ret))
        cls.restore_serial_timeout(instrument)
        return ret

    @classmethod
    def write_ram(cls, instrument, regs_row):
        """
        Writing a sequence of holding regs from RAM (RAM_BOUNDS[1] - RAM_BOUNDS[0])
        """
        regs_length = len(regs_row)
        if regs_length > cls.RAM_BOUNDS[1] - cls.RAM_BOUNDS[0]:
            raise RuntimeError("Too long sequence! (should not be more %d)" % cls.RAM_BOUNDS[1] - cls.RAM_BOUNDS[0])
        logging.debug("Writing signal to RAM as %d regs with %.2fs serial timeout..." % (regs_length, cls._SERIAL_TIMEOUT))
        cls.set_serial_timeout(instrument, cls._SERIAL_TIMEOUT)
        _max_len = 122  # for modbus funcode 6
        if regs_length > _max_len:  # will write _max_len regs once
            instrument.write_u16_regs(cls.RAM_BOUNDS[0], regs_row[:_max_len])
            _beginning = cls.RAM_BOUNDS[0] + _max_len
            for reg in regs_row[_max_len:]:
                instrument.write_u16(_beginning, reg)
                _beginning += 1
        elif regs_length == 0:
            logging.warning("Will not write signal")
        else:
            instrument.write_u16_regs(cls.RAM_BOUNDS[0], regs_row)
        cls.restore_serial_timeout(instrument)

    @classmethod
    def in_rom(cls, instrument, bank, callable):
        instrument.write_bit(cls.EDIT_BANK_BASE + bank, True)
        logging.debug("Turned ROM bank %d to editing" % bank)
        time.sleep(cls._SCHEDULER_PERIOD)
        ret = callable()
        instrument.write_bit(cls.EDIT_BANK_BASE + bank, False)
        logging.debug("Stopped editing of ROM bank %d" % bank)
        time.sleep(cls._SCHEDULER_PERIOD)
        return ret

    @classmethod
    def read_rom(cls, instrument, bank):
        regs_length = instrument.read_u16(cls.BANK_BYTELENGTH_BASE + bank)
        logging.debug("ROM bank %d has %d ir regs" % (bank, regs_length))
        return cls.in_rom(instrument, bank, lambda: cls.read_ram(instrument, regs_length))

    @classmethod
    def write_rom(cls, instrument, bank, regs_row):
        logging.debug("Writing signal to %d ROM bank\n%s" % (bank, str(regs_row)))
        callable = lambda: cls.write_ram(instrument, regs_row)
        return cls.in_rom(instrument, bank, callable)


def backup_ir(instrument):
    """
    Reading IR cmds from all banks and writing to a file
    """
    result = {}
    ir_base = IRMixin()
    result_fname = ir_base.BACKUP_FNAME % instrument.get_serial_number()
    ir_base.clear_ram(instrument)
    banks_to_read, _ = ir_base.get_used_banks(instrument)

    logging.debug("Will read IR CMDs from banks %s to %s" % (str(banks_to_read), result_fname))

    for bank in banks_to_read:
        ret = ir_base.read_rom(instrument, bank)
        result.update({'%d' % bank : ret})
        beeper.beep(0.15, 1)

    with open(result_fname, 'wt') as fp:
        fp.write(str(result))


def restore_ir(instrument, backup_fname):
    """
    Restoring backup from latest fname, containing actual sn
    """
    # backup_files = glob.glob(os.path.join(BACKUP_DIR, '*%d*' % instrument.get_serial_number()))
    # latest_backup = max(backup_files, key=os.path.getctime)
    # logging.debug("Found latest backup file %s" % latest_backup)
    written = 0

    backup = literal_eval(open(backup_fname).read())

    ir_base = IRMixin()
    ir_base.clear_ram(instrument)

    logging.debug("Will write backed up IR CMDs from %s to device" % backup_fname)

    for bank, ir_cmd in backup.items():
        ir_base.write_rom(instrument, int(bank), ir_cmd)
        beeper.beep(0.15, 1)
        written += 1

    return written
